/*  
    Created by Neil Killeen
    4/16/14
 */

#define _POSIX_SOURCE
#include <stdio.h>
#include <unistd.h>
#include <sys/types.h>
#include <errno.h>

#include <string.h>
#include <stdlib.h>
#include <sys/wait.h>
#include <signal.h>

extern char **get_line(void);

typedef enum {
    IN = 0x1,
    OUT = 0x1 << 1,
    PIPE = 0x1 << 2,
    NONE = 0x0
} Redirection;

typedef struct {
    char *out_file;
    char *in_file;
} FileStruct;

/*  
    Counts the number of elements in a 2D char *
    @param  args    A set of tokens

    @return Number of elements in args
*/
int count(char **args) 
{
    int i;
    for (i = 0; args[i] != NULL; ++i) {}
    return i;
}

/*  
    Pops the head token and returns it
    @param  args    A set of tokens

    @return A pointer to the old head token
 */
char *pop_token(char ***args) 
{
    char *element;
    if ((*args)[0] != NULL) {
        element = (*args)[0];
        *args = &((*args)[1]);
        return element;
    }
    return (char *)NULL;
}

/*  
    Pops the head token and frees it.
    @param  args    A set of tokens
*/
void pop_token_free(char ***args) 
{
    char *popped = pop_token(args);
    free(popped);
    popped = NULL;
}

/*
    Gets the last token in a set.
    @param  args    A set of tokens

    @return A pointer to the current tail token
*/
char *tail(char **args) 
{
    int index = count(args) - 1;
    return args[index];
}

/*
    Frees the memory to the tail token.
    @param  args    A set of tokens
*/
void free_tail(char ***args) 
{
    int index = count(*args) - 1;
    if ((*args)[index] != NULL) {
        free((*args)[index]);
        (*args)[index] = NULL;
    }
}

/*
    Prints the shell prompt.
*/
void prompt(void) {
    printf("@ ");
}

/*
    Checks to see if the the parent should wait.
    Frees the token associated with the wait.
    @param  args    A set of tokens

    @return [0] if the parent can continue
            [1] if the parent must wait
*/
int is_waiting(char **args) {
    int index = count(args) - 1;
    if (!strcmp(args[index], "&")) {
        free(args[index]);
        args[index] = NULL;
        return 0;
    }
    return 1;
}

/* 
    Finds a token in a set of tokens.
    @param  args        A set of tokens
    @param  token       Token to search for in a set of tokens

    @return [-1]        If it couldn't find the token, 
            [position]  If it could
*/
int find_token(char **args, char *token) {
    int i;
    for (i = 0; args[i] != NULL; ++i) {
        if (!strcmp(args[i], token)) return i;
    }
    return -1;
}

/*
    Remove a redirection from a set of tokens,
    set the file reference, and rearrange the set of tokens.
    @param  args    A set of tokens
    @param  file    The filename which will be set to a removed token
    @param  offset  Where the redirection started

    @return [-1]    If there is an error
            [0]     If the function succeeds.
*/
int rm_file_tokens(char ***args, char **file, int offset) {
    int i;
    *file = (*args)[offset + 1];
    if (*file == NULL) {
        perror("Syntax error, no file name.\n");
        return -1;
    }

    free((*args)[offset]);
    (*args)[offset] = NULL;
    (*args)[offset + 1] = NULL;

    for (i = 0; (*args)[offset + 2 + i] != NULL; ++i) {
        (*args)[offset + i] = (*args)[offset + 2 + i];
        (*args)[offset + 2 + i] = NULL;
    }
    return 0;
}

/*
    Check if there are redirections in a set of tokens.
    @param  args        A set of tokens
    @param  files       A structure of input and output filenames

    @return [IN]        Only input redirection
            [OUT]       Only output redirection
            [IN + OUT]  Both input and output redirection
*/
Redirection check_redir(char ***args, FileStruct *files) {
    int errno;
    int outcheck;
    int incheck = find_token((*args), "<");
    Redirection redir = NONE;
    
    if (incheck != -1) /* File in */ {
        errno = rm_file_tokens(args, &(*files).in_file, incheck);
        if (!errno)
            redir += IN;
    }

    outcheck = find_token((*args), ">");
    if (outcheck != -1) /* File out */ {
        errno = rm_file_tokens(args, &(*files).out_file, outcheck);
        if (!errno)
            redir += OUT;
    }

    return redir;
}

/*
    Callback for the sigaction_handler.
    Calls wait to reap any zombie children.
    @param  signal  Signal number 
*/
void sig_handler(int signal) {
    int status;
    pid_t pid;
    if (signal == SIGCHLD) {
        pid = waitpid(-1, &status, 0);
        if (pid == -1)
            perror("Error waiting for child.\n");
    }
}

/*
    Initializes the signal handler and sets
    all flags to 0.
 */
void init_handler(void) {
    struct sigaction new_action, old_action;
    int ret;

    new_action.sa_handler = sig_handler;
    ret = sigemptyset(&new_action.sa_mask);
    if (ret == -1)
        perror("Error in sigemptyset\n");

    new_action.sa_flags = 0;
    ret = sigaction(SIGCHLD, NULL, &old_action);
    if (ret == -1)
        perror("Error in sigaction\n");

    if (old_action.sa_handler != SIG_IGN) {
        ret = sigaction(SIGCHLD, &new_action, NULL);
        if (ret == -1)
            perror("Error in sigaction\n");
    }
}

int main(int argc, char *argv[]) {
    int i;
    char **args;
    int status, ret, wait;
    pid_t child_pid;
    FileStruct files;
    Redirection redir;

    files.out_file = NULL;
    files.in_file = NULL;
    
    init_handler();
    
    for(;;) {
        prompt();
        args = get_line();
        
        if (!args) break;
        if (!args[0]) continue;

        wait = is_waiting(args);
        redir = check_redir(&args, &files);
        
        /* Exit gracefully if asked to exit */
        if (!strcmp(args[0], "exit")) 
            _exit(EXIT_SUCCESS);
        
        /* Execute process */
        child_pid = fork();
        
        switch (child_pid) {
            case -1: { /* Error */
                perror("Forking error.\n");
                exit(EXIT_FAILURE);
                break;
            }
            case 0: { /* Child */
                switch (redir) {
                    case (IN + OUT):
                        if (!freopen(files.in_file, "r", stdin)) {
                            perror("Error redirecting from file\n");
                            redir = redir & (~IN);
                        }
                        if (!freopen(files.out_file, "w", stdout)) {
                            perror("Error redirecting to file\n");
                            redir = redir & (~OUT);
                        }
                        ret = execvp(args[0], args);
                        fclose(stdout);
                        fclose(stdin);

                        free(files.in_file);
                        free(files.out_file);
                        break;
                    case IN:
                        if (!freopen(files.in_file, "r", stdin)) {
                            perror("Error redirecting from file\n");
                            redir = redir & (~IN);
                        }
                        ret = execvp(args[0], args);
                        fclose(stdin);

                        free(files.in_file);
                        break;
                    case OUT:
                        if (!freopen(files.out_file, "w", stdout)) {
                            perror("Error redirecting to file\n");
                            redir = redir & (~OUT);
                        }
                        ret = execvp(args[0], args);
                        fclose(stdout);

                        free(files.out_file);
                        break;
                    default:
                        ret = execvp(args[0], args);
                        break;
                }

                files.in_file = NULL;
                files.out_file = NULL;
                if (ret == -1)
                    perror("Error executing the command.\n");
                break;
            }
            default: { /* Parent */
                if (wait)
                    ret = waitpid(-1, &status, 0);
                if (ret == -1)
                    perror("Error waiting for child.\n");
                break;
            }
        }
    }
    return 0;
}
